# ============================================================================ #

KolmogorovDistance2Unif <- function(x,Fx) max(abs(x - Fx))

# ============================================================================ #

LevyDistance2Unif <- function(x, Fx) {
  return( sqrt( sum( x^2 + Fx^2 -(x^2+x*Fx) - (x*Fx- Fx^2) + (x+Fx)^2/2 ) ) )
}

# ============================================================================ #
#' Tests adherence to the uniform distribution
#'
#' @author Rodrigo Labouriau
#' @param x a vector with the values
#' @param Nboots number of bootstrap simulations
#' @param DrawPlot should a plot of the CDF drawn (default = FALSE)
#' @param QQplot should a QQ-plot be drawn (default = FALSE)
#' @return the p-value of the test.
#' @details Tests the adherence to the uniform distribution by calculating the
#' Kolmogorov distance between the empirical cumulative distribution function
#' and the cumulative distribution function of a uniform distribution on (0,1),
#' and using a bootstrap test for testing equality of the distance to zero.
#' @examples TestUnif(runif(50))
#' @export
TestUnif <- function(x, Nboots = 9999, DrawPlot = FALSE, QQplot = FALSE){
  x <- sort(x)
  n <- length(x)
  vals <- unique(x)
  Fval <- cumsum(tabulate(match(x, vals)))/n
  KD <- KolmogorovDistance2Unif(vals, Fval)
  BootKD <- numeric(Nboots)
  for(b in 1:Nboots){
    Xb <- runif(n)
    Xb <- sort(Xb)
    valsB <- unique(Xb)
    FvalB <- cumsum(tabulate(match(Xb, valsB)))/n
    BootKD[b] <- KolmogorovDistance2Unif(valsB, FvalB)
  }
  Pvalue <- round(mean(KD< BootKD), digits = ceiling(log10(Nboots)))
  if(DrawPlot){
    plot(vals, Fval, type="s", lwd=2, col="blue", xlab = "Value", ylab = "ECDF",
         sub = paste(
          "P-value for adherence to the U(0,1) distribution =", Pvalue)
    )
    abline(0,1)
  }
  if(QQplot){
    QQunif(x,
           sub = paste("P-value for adherence to the U(0,1) distribution =",
                        Pvalue))
  }
  return(Pvalue)
}

# ============================================================================ #

# ============================================================================ #
QQunif <- function (x,
  ReturnValues = FALSE,
  DrawPlot = TRUE,
  Add = FALSE,
  main = "",
  sub = "")
{
  x <- sort(x)
  n <- length(x)
  if (n < 1)
    stop("'x' must have 1 or more non-missing values")
  vals <- unique(x)
  rval <- cumsum(tabulate(match(x, vals)))/n
  Out <- list(Observed=vals, Theoretical=rval)
  if(DrawPlot){
    if(!Add){
      plot(Out[[2]], Out[[1]], cex=0.8, col="gray", pch=19,
        ylim=c(0,1),
        xlab="Theoretical Uniform Quantiles",
        ylab="Observed Uniform Quantiles",
        main = main,
        sub = sub)
      abline(0,1)
      points(Out[[2]], Out[[1]], cex=0.8)
    }
    if(Add){
      points(Out[[2]], Out[[1]], type="l")
    }
  }
  if(ReturnValues) return(Out)
}

# ============================================================================ #

# ============================================================================ #
UtransfGamma <- function(y, mu, Scale){
  N <- length(y)
  U <- numeric(N)
  for(i in 1:N){
    U[i] <- pgamma(q=y[i], shape=mu[i], scale=Scale)
  }
  return(U)
}
# ============================================================================ #

# ============================================================================ #
TestGLMgamma <- function(
  Model,                    # a model of class glm or glmerMod
  ParametricBoots = NULL,   # number of parametric bootstrap samples (Default = NULL, no parametric bootstrap is done)
  PlotScale = FALSE,        # should the scale search plot be drawn
  QQplot = TRUE,            # should the uniform QQ-plot be drawn (including p-values as sub-title)
  QQplotSimple = FALSE,     # should a simplified qq-plot (without p-values) be drawn
  ReturnTransform = TRUE,   # should uniform transformed be returned (Default = TRUE)
  MaxScale = NULL,          # maximum value of the scale in a grid search (if Null, sum of Pearson residuals)
  MinScale = NULL,          # minimum value of the scale in a grid search (if Null, stepSearch)
  stepSearch = NULL,        # number of grid searchs for finding the scale yielding the minimum Kolmogorov distance for the scale (Default = NULL, no grid search is performed). This is slow.
  Nboots = 9999,            # number of bootstrap samples for testing uniformity
  Scale = NULL,             # supplied value of the scale (Defaut = NULL, no value is supplied)
  PvalueGridSearch = FALSE, # should find the scale that miximizes the p-value (Default = NULL, minimize the Kolmogorov distance)
  Seed = NULL,              # seed for the random sequence used in the test of uniformity
  Coverage = 0.95,          # coverage of the confidence band for parametric bootstrap
  ModelBoot = TRUE          # should the parametric bootstrap consider fitted variation (Default = TRUE)
){

  if (class(Model)[1] != "glmerMod" & class(Model)[1] != "glm"){
    stop("Please, the model should be of class glm or glmerMod")
  }

  if(class(Model)[1] == "glmerMod") y <- Model@resp$y
  if(class(Model)[1] == "glm") y <- Model$y

  Fitted <- fitted(Model)

  if(is.null(Scale)){
    # When the Scale is NOT given

    # Grid search of Scale via minimization of the Kolmogorov distance
    if (!is.null(stepSearch) & !PvalueGridSearch){
      ff <- function(x){
        UU <- UtransfGamma(y = y, mu = Fitted/x, Scale = x)
        UU <- sort(UU)
        n <- length(UU)
        vals <- unique(UU)
        Fval <- cumsum(tabulate(match(UU, vals)))/n
        Distance <- KolmogorovDistance2Unif(vals, Fval)
        return(Distance)
      }

      if(is.null(MaxScale)) MaxScale <- sum(abs(residuals(Model, "pearson")))
      if(is.null(MinScale)) MinScale <- stepSearch
      Grid <- seq(from = MinScale, by = stepSearch, to = MaxScale)

      FF <- numeric(length(Grid))
      for(i in 1:length(Grid)){
        FF[i] <- ff(Grid[i])
      }

      Scale <- Grid[which.min(FF)]

      if(PlotScale){
        plot(Grid, FF, type = "l",
             xlab = "Scale",
             ylab = "Kolmogorov Distance to Uniform",
             sub = paste("Scale estimate =", Scale))
        abline(v = Scale, lty = 2, lwd = 0.2)
      }
    }

    # Grid search of Scale via maximization of the pvalue
    if (!is.null(stepSearch) & PvalueGridSearch){

      if(is.null(MaxScale)) MaxScale <- sum(abs(residuals(Model, "pearson")))
      if(is.null(MinScale)) MinScale <- stepSearch
      Grid <- seq(from = MinScale, by = stepSearch, to = MaxScale)

      Pv <- numeric(length(Grid))
      a <- 0
      for(s in Grid){
        a <- a + 1
        UU <- UtransfGamma(y = y, mu = Fitted/s, Scale = s)
        Pv[a] <- TestUnif(x = UU, Nboots = Nboots,
                          QQplot = FALSE, DrawPlot = FALSE)
      }

      Scale <- Grid[which.max(Pv)]

      if(PlotScale){
        plot(Grid, Pv, type = "l",
             xlab = "Scale",
             ylab = "P-value",
             sub = paste("Scale estimate =", Scale))
        abline(v = Scale, lty = 2, lwd = 0.2)
      }

      # plot(GRID, Pv)
      # abline(v = Scale, lty = 2)
    }
  }

  # END of the procedure when a Scale search is performed

  UU <- UtransfGamma(y = y, mu = Fitted/Scale, Scale = Scale)
  if(!is.null(Seed)) set.seed(Seed)
  Pvalue <- TestUnif(x = UU, Nboots = Nboots, QQplot = QQplot, DrawPlot = FALSE)

  if(QQplotSimple) qq_unif(x = UU, Add = FALSE)

  # When parametric bootstrap is required
  if (!is.null(ParametricBoots)){
    ##
    Simulate <- function(Fitted, Scale){
      Fitted <- fitted(Model)
      SIM <- rgamma(n = length(Fitted), shape = Fitted/Scale, scale = Scale)
      return(SIM)
    }
    ##
    BootUU <- matrix(nrow=ParametricBoots, ncol=length(UU))

    if(class(Model)[1] == "glm") Dboot <- Model$data
    if(class(Model)[1] == "glmerMod") Dboot <- Model@frame

    Formula <- formula(Model)
    Family <- family(Model)
    for (b in 1:ParametricBoots){
      yboot <- Simulate(Fitted = Fitted, Scale = Scale)
      if(ModelBoot){ # This is SLOW when using models with random components
        Dboot[ ,as.character(Formula[2])] <- yboot
        BootModel <- update(Model, data = Dboot)
        BootFitted <- fitted(BootModel)
        UUboot <- UtransfGamma(y = yboot, mu = BootFitted/Scale, Scale = Scale)
      } else {
        UUboot <- UtransfGamma(y = yboot, mu = Fitted/Scale, Scale = Scale)
      }
      BootUU[b, ] <- UUboot
    }

    UpperUU <- LowerUU <- numeric(length(UU))
    for(i in 1:(length(UU))){
      LowerUU[i] <- quantile(BootUU[ ,i], probs = (1-Coverage)/2)
      UpperUU[i] <- quantile(BootUU[ ,i], probs = (1 - (1-Coverage)/2))
    }

    LowerUU <- sort(LowerUU)
    UpperUU <- sort(UpperUU)


    qq_unif(x = UU, Add = FALSE)
    GG <- (1:length(UU))/length(UU)
    # points(GG, LowerUU, type = "l", lwd = 2)
    # points(GG, UpperUU, type = "l", lwd = 2)

    polygon(c(UpperUU,rev(LowerUU)), c(GG,rev(GG)), col="aliceblue")

    qq_unif(x = UU, Add = TRUE)
    abline(0,1, lty = 2)

  }
  # END of the parametric bootstrap module

  if(!ReturnTransform) UU <- NULL

  Out <- list(p.value = Pvalue, scale = Scale, transformedFit = UU)

  return(Out)
}
# ============================================================================ #

# ============================================================================ #
qq_unif <- function (x, Add = FALSE){
  x <- sort(x)
  n <- length(x)
  if (n < 1)
    stop("'x' in qq_unif must have 1 or more non-missing values")
  vals <- unique(x)
  rval <- cumsum(tabulate(match(x, vals)))/n
  Out <- list(Observed=vals, Theoretical=rval)
  if(!Add){
    plot(Out[[2]], Out[[1]], type="l", lwd = 3, col = "red",
         ylim=c(0,1),
         xlab="Theoretical Uniform Quantiles",
         ylab="Observed Uniform Quantiles")
    abline(0,1, lty = 2)
  }
  if(Add){
    points(Out[[2]], Out[[1]], type="l", lwd = 3, col = "red")
    abline(0,1, lty = 2)
  }
}

# ============================================================================ #
# ============================================================================ #

# ============================================================================ #
MonteCarloModelTest <- function(Model, Nsim=1000){

  if(class(Model)[1] == "glmerMod") y <- Model@resp$y
  if(class(Model)[1] == "glm") y <- Model$y

  KD <- KolmDist(y, fitted(Model))

  SimKDs <- numeric(Nsim)
  for(b in 1:Nsim){
    SimKDs[b] <- KolmDist(simulate(object=Model, nsim=1)[[1]] ,
      fitted(Model) )
  }
  Pvalue <- mean(KD<SimKDs, na.rm = TRUE)

  return(Pvalue)
}
# ============================================================================ #

# ============================================================================ #
KolmDist <- function(x, y){
  z <- c(x,y); z <- sort(z)
  Nx <- length(x); Ny <- length(y)
  SFG <- unique(z)
  N <- length(SFG)

  Delta <- numeric(N)
  for(i in 1:N){
    Delta[i] <- abs( ((length(x[x <= SFG[i]]))/ Nx ) -
        ((length(y[y <= SFG[i]]))/ Ny )
    )
  }
  return(max(Delta))
}
# ============================================================================ #

