# ============================================================================ #
#' Calculates Bootstrap p-values for a Chi-Square Goodness of Fit Test
#'
#' @author Rodrigo Labouriau
#' @param Expected a vector with the expected values.
#' @param Observed a vector with the observed values (same dimension as the
#' expected).
#' @param Nsim the number of bootstrap simulations (default = 1000).
#' @return a (numeric) p-value
#' @details Calculates a non-parametric bootstrap estimate of the p-value
#' of chis-quare goodness of fit test.
#' @examples x <- rpois(20, 5);  K <- max(x, na.rm=TRUE) + 1
#' Expected <- Observed <- numeric(K) for(k in 0:K){
#'  Expected[k] <- sum( dpois(x=k-1, lambda=lambda) )
#'  Observed[k] <- sum( x==k-1 )
#'  }
#'  SimPvalues (Expected, Observed, Nsim=999)
#' @export
SimPvalues <- function
( Expected,
  Observed,
  Nsim=1000
){
  K <- length(Observed)
  Nobs <- sum(Observed)
  Prob <- Expected/sum(Expected)
  Chsq <-  sum ( ( Observed-Expected )^2 /Expected )
  # Pvalue.asymptotic <- pchisq(q=Chsq, df=(K-1), lower.tail=FALSE)
  SimChisquares <- numeric(Nsim)
  for(i in 1:Nsim){
    Counts <- sample(x=1:K, prob=Prob, size=Nobs, replace=TRUE)
    SimObserved <- numeric(K)
    for(k in 1:K){
      SimObserved[k] <- sum( Counts==k )
    }
    SimChisquares[i] <-  sum ( ( SimObserved-Expected )^2 /Expected )
  }
  return(mean(Chsq<SimChisquares))
}
# ==============================================================================

# ==============================================================================
BootCIobserved <- function
( Expected,
  Observed,
  Nsim = 1000,
  Coverage = 0.95,
  MultipleCompAdjust = NULL
){
  K <- length(Observed)
  Nobs <- sum(Observed)
  Prob <- Expected/sum(Expected)
  SimObserved <- matrix(nrow=Nsim, ncol=K)
  for(i in 1:Nsim){
    Counts <- sample(x=1:K, prob=Prob, size=Nobs, replace=TRUE)
    for(k in 1:K){
      SimObserved[i, k] <- sum( Counts==k )
    }
  }
  Lower <- Upper <- Pvalues <- numeric(K)
  for(k in 1:K){
    Lower[k] <- quantile(SimObserved[, k], probs=((1-Coverage)/2) )
    Upper[k] <- quantile(SimObserved[, k], probs=(1-(1-Coverage)/2) )
    Pless <- mean(Observed[k]<= SimObserved[, k])
    Pmore <- mean(Observed[k] >= SimObserved[, k])
    Pvalues[k] <- min(Pless, Pmore)
  }
  if(!is.null(MultipleCompAdjust)){
    Pvalues <- p.adjust(Pvalues, method = MultipleCompAdjust)
  }
  return( cbind(Lower, Upper, Pvalues) )
}
# ==============================================================================

# ============================================================================ #
#' Method for printing objects of class testDiscreteDistribution
#'
#' @author Rodrigo Labouriau
#' @param x an object of class testDiscreteDistribution.
#' @param digits number of digits for printing (default = 4)
#' @param ... further arguments passed to or from other methods
#' @return nothing
#' @details
#' @examples
#' @export
print.testDiscreteDistribution <- function (x, digits = 4, ...){
  print(round(x$Pvalue, digits=digits))
}
# ==============================================================================

# ==============================================================================
summary.testDiscreteDistribution <- function
( testDiscreteDistributionObj,
  Nrows=10,
  Nsim=1000,
  Coverage=0.95,
  digits=3,
  MultipleCompAdjust = NULL
){

  Model <- testDiscreteDistributionObj$Model
  if (class(Model)[1] == "glmerMod"){
    Call <- as.character( Model@call )
  } else if (class(Model)[1] == "glm" |
             class(Model)[1] == "lm"  |
             class(Model)[1] ==  "negbin"){
    Call <- as.character( Model$call )
  } else stop("Model should be of class glmerMod, glm or lm.")

  cat("Call:", Call, "\n")

  cat("P-value = ", testDiscreteDistributionObj$Pvalue, "\n")

  CI <- BootCIobserved(Expected = testDiscreteDistributionObj$Expected,
                       Observed = testDiscreteDistributionObj$Observed,
                       Nsim = Nsim,
                       Coverage = Coverage,
                       MultipleCompAdjust = MultipleCompAdjust)

  MM <- cbind(1:length(testDiscreteDistributionObj$Observed)-1,
              testDiscreteDistributionObj$Expected,
              testDiscreteDistributionObj$Observed,
              CI
    )
  colnames(MM) <- c("Value", "Exp.Freq.", "     Obs.Freq.",
                     paste("L.Lim.(", (1-Coverage)/2 *100, "%)", sep=""),
                     paste("U.Lim.(",100-(1-Coverage)/2 *100, "%)", sep=""),
                     "p-value"
                    )

  MM <- MM[ , c(1,2,4,3,5,6)]

  row.names(MM) <- rep("", dim(MM)[1])
  Nrows <- min(Nrows, dim(MM)[1])
  MM <- round(MM, digits = digits)
  print(MM[1:Nrows , ])

  if(Nrows < dim(MM)[1]) cat("... (truncated table)")

}
# ==============================================================================

# ==============================================================================
plot.testDiscreteDistribution <- function
( testDiscreteDistributionObj,
  Nsim=1000,
  Coverage=0.95,
  Main="",
  Sub="",
  CalculateCI=TRUE,
  ScatterPlot=FALSE
){

  Model <- testDiscreteDistributionObj$Model

  if (class(Model)[1] == "glmerMod"){
    Formula <- Model@call$formula
    Data <- Model@frame
    Family <- Model@resp$family
    Response <- as.character( Formula[[2]] )
    ResponsePosition <- which(names(Data)== Response)
  }

  if (class(Model)[1] == "glm" ){
    Formula <- formula(Model)
    Data <- Model$model
    Family <- Model$family
    Response <- as.character( Formula[[2]] )
    ResponsePosition <- which(names(Data)== Response)
  }

  if (class(Model)[1] == "negbin" ){
    Formula <- formula(Model)
    Data <- Model$model
    Family <- Model$family
    Response <- as.character( Formula[[2]] )
    ResponsePosition <- which(names(Data)== Response)
  }

  Counts <- Data[[ResponsePosition]]
  Fitted <- fitted(Model)
  K <- max(Counts, na.rm=TRUE) + 1
  Expected <- Observed <- numeric(K)
  for(k in 0:K){
    Expected[k] <- sum( dpois(x=k-1, lambda=Fitted) )
    Observed[k] <- sum( Counts==k-1 )
  }

  if(CalculateCI)
  {
    ObservedSim <- matrix(nrow=Nsim, ncol=K)
    for(s in 1:Nsim){
      Data[[ResponsePosition]] <- simulate(Model)[[1]]

      if (class(Model)[1] == "glmerMod"){
        suppressWarnings(
         tryCatch({
            ModelSim <- glmer(formula=Formula, family=poisson, data=Data)
            CountsSim <- ModelSim@resp$y
           }, error=function(e){}
         )
        )
      }

      if (class(Model)[1] == "glm"){
        suppressWarnings(
          tryCatch({
            ModelSim <- glm(formula=Formula, family=Family, data=Data)
            CountsSim <- ModelSim$y
          }, error=function(e){}
         )
       )
      }

      if (class(Model)[1] == "negbin"){
        ModelSim <- glm.nb(formula=Formula, data=Data)
        CountsSim <- ModelSim$y
      }

      for(k in 0:K){
        ObservedSim[s,k] <- sum( CountsSim==k-1 )
      }

    }
    Lower <- Upper <- numeric(K)
    for(k in 1:K){
      Lower[k] <- quantile(ObservedSim[, k], probs=((1-Coverage)/2) )
      Upper[k] <- quantile(ObservedSim[, k], probs=(1-(1-Coverage)/2) )
    }
  }

 if( ScatterPlot ){                      # Drawing a scatter plot

  Symbol <- as.character( 0:(K-1) )
  plot(Observed, Expected, pch=Symbol, main=Main, sub=Sub)
  abline(0,1)
  if(CalculateCI){
    plot(Observed, Expected, pch=Symbol, main=Main, sub=Sub,
         xlim=c(min(Lower), max(Upper)))
    abline(0,1)
    points(Lower, Expected, pch="|", col="gray")
    points(Upper, Expected, pch="|", col="gray")
  }
 } else{                                # Drawing a histogram
  plot(0:(K-1)+0.25, Expected, type="h", xlab="", ylab="", lwd=6, col="red",
    ylim=c(0,max( max(Upper), max(Observed), max(Expected)) ), xlim=c(0,K),
    sub=Sub, main=Main)
  points(0:(K-1), Observed, type="h", xlab="", ylab="", lwd=6,col="blue")
  legend(x="topright", legend=c("Expected","Observed"), col=c("red", "blue"),
    lwd=6, bty="n")
  if(CalculateCI){
    points(0:(K-1), Lower, type="b", lwd=2, lty=3)
    points(0:(K-1), Upper, type="b", lwd=2, lty=3)
  }
 }
}
# ==============================================================================

# ==============================================================================
scatterPlot.testDiscreteDistribution <- function
( testDiscreteDistributionObj,
  Main="",
  Sub=""
){

  Expected <- testDiscreteDistributionObj$Expected
  Observed <- testDiscreteDistributionObj$Observed

  K <- length(Observed)

  Symbol <- as.character( 0:(K-1) )
  plot(Expected, Observed, pch=Symbol, main=Main, sub=Sub)
  abline(0,1)

}
# ==============================================================================

# ============================================================================ #

BootOverDispersion <- function(Model, Nboots=1000, Coverage=0.95){

  if(!("glm" %in% class(Model) & (Model$family[1] == "quasipoisson" | Model$family[1] == "quasibinomial"))){
    stop("The model should be of class glm and the family quasipoisson or quasibinomial")
  }

 Data <- model.frame(Model)

 Dispersion <- summary(Model)$dispersion
 Nobs <- nobs(Model)
 DispersionBoots <- numeric(Nboots)
 for(b in 1:Nboots){
    BootData <- Data[sample(x=1:Nobs, size=Nobs, replace = TRUE) , ]
    Mboot <- update(Model, data=BootData)
    DispersionBoots[b] <- summary(Mboot)$dispersion
  }
  CI <- quantile(x = DispersionBoots, probs = c((1-Coverage)/2, 1-(1-Coverage)/2))
  Pvalue <- mean(DispersionBoots <= 1)
  Out <- list(CI = CI,
              Pvalue = Pvalue,
              Dispersion = Dispersion,
              BootstrapDispersions = DispersionBoots,
              Coverage = Coverage
            )
  return(Out)
}

# ============================================================================ #
